"""Defines the base event class for all subclasses of events."""

from __future__ import annotations

from dataclasses import dataclass

from eth_typing import ChecksumAddress
from hexbytes import HexBytes


@dataclass(kw_only=True)
class BaseEventArgs:
    """The base event argument class for all subclasses of event args."""


@dataclass(kw_only=True)
class BaseEvent:
    """The base event class for all subclasses of events."""

    # pylint: disable=too-many-instance-attributes

    # Base event arguments
    log_index: int
    transaction_index: int
    transaction_hash: HexBytes
    address: ChecksumAddress
    block_hash: HexBytes
    block_number: int
    args: BaseEventArgs
    __name__: str = "BaseEvent"



"""Dataclasses for all structs in the Example contract.

DO NOT EDIT.  This file was generated by pypechain v0.0.46.
See documentation at https://github.com/delvtech/pypechain """

# super() call methods are generic, while our version adds values & types
# pylint: disable=arguments-differ

# contracts have PascalCase names
# pylint: disable=invalid-name

# contracts control how many attributes and arguments we have in generated code
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments

# unable to determine which imports will be used in the generated code
# pylint: disable=unused-import

# we don't need else statement if the other conditionals all have return,
# but it's easier to generate
# pylint: disable=no-else-return

# We import this contract itself to ensure all nested structs have a fully qualified name.
# We use this to avoid namespace collisions, as well as having a uniform
# type structure to do lookups when functions return these structs.
# pylint: disable=import-self


from __future__ import annotations

from dataclasses import dataclass

from pypechain.core import BaseEvent, BaseEventArgs, ErrorInfo, ErrorParams

from ..Example import ExampleTypes as Example


@dataclass(kw_only=True)
class FlipEvent(BaseEvent):
    """The event type for event Flip"""

    @dataclass(kw_only=True)
    class FlipEventArgs(BaseEventArgs):
        """The args to the event Flip"""

        flip: int

    # We redefine the args field with the specific event arg type.
    args: FlipEventArgs  # type: ignore[override]

    __name__: str = "Flip"


@dataclass(kw_only=True)
class FlopEvent(BaseEvent):
    """The event type for event Flop"""

    @dataclass(kw_only=True)
    class FlopEventArgs(BaseEventArgs):
        """The args to the event Flop"""

        flop: int

    # We redefine the args field with the specific event arg type.
    args: FlopEventArgs  # type: ignore[override]

    __name__: str = "Flop"


@dataclass
class SimpleStruct:
    """SimpleStruct struct."""

    intVal: int
    strVal: str


@dataclass
class InnerStruct:
    """InnerStruct struct."""

    boolVal: bool


@dataclass
class NestedStruct:
    """NestedStruct struct."""

    intVal: int
    strVal: str
    innerStruct: Example.InnerStruct


WrongChoiceError = ErrorInfo(
    inputs=[
        ErrorParams(name="answer", python_type="int", solidity_type="uint8"),
        ErrorParams(name="errorMessage", python_type="str", solidity_type="string"),
    ],
    name="WrongChoice",
    selector="0xc13b30d4",
    signature="WrongChoice(uint8,string)",
)



"""Script to recursively format json files under a directory."""

import json
import os


def _format_json_dir(json_dir):
    for root, dirs, files in os.walk(json_dir):
        # Format any outer json files
        for file in files:
            if file.endswith(".json"):
                json_file = os.path.join(root, file)
                # Format the JSON file
                with open(json_file, "r", encoding="utf-8") as file:
                    data = json.load(file)
                with open(json_file, "w", encoding="utf-8") as file:
                    json.dump(data, file, ensure_ascii=False, indent=2)

        for d in dirs:
            _format_json_dir(d)


if __name__ == "__main__":
    _format_json_dir("example/abis/")
    _format_json_dir("pypechain/test/")



"""Export all types from generated files.

DO NOT EDIT.  This file was generated by pypechain v_version_ignored.
See documentation at https://github.com/delvtech/pypechain """

# The module name reflects that of the solidity contract,
# which may not adhere to python naming conventions
# pylint: disable=invalid-name


from .ContractFile1 import ContractFile1Contract


from .ContractFile2 import ContractFile2Contract


from .TypesFile import Struct1


from .TypesFile import Struct2




"""The contract function base class for pypechain."""

from web3.contract.contract import ContractFunction


class PypechainContractFunction(ContractFunction):
    """The contract function base class for pypechain."""

    # The function name for the contract function
    _function_name: str

    # The python type signatures for the contract function
    _type_signature: str



"""Tests for overloading methods."""

from __future__ import annotations

import pytest
from web3.exceptions import ContractCustomError

from pypechain.core import PypechainCallException

from .types.Errors import Ages, ErrorsContract


@pytest.mark.usefixtures("process_contracts")
class TestErrors:
    """Tests events emitted from the contracts."""

    def test_error_one(self, w3):
        """Test that we can decode just the selector."""
        deployed_contract = ErrorsContract.deploy(w3=w3, account=w3.eth.accounts[0])

        # Test error handling with "call"

        # Calls should fail and caught with custom error
        try:
            _ = deployed_contract.functions.revertWithErrorOne().call()
            assert False, "Expected exception."
        except PypechainCallException as err:
            # Catch custom error
            assert isinstance(err.orig_exception, ContractCustomError)
            assert err.decoded_error == "One()"
            assert err.decoded_error_name == "One"
            assert err.decoded_error_args == ()
            assert err.contract_call_type == "call"
            assert err.function_name == "revertWithErrorOne"
            assert err.fn_args == ()
            assert err.fn_kwargs == {}
            # Call functions don't write a block, so the block should be identical
            assert err.block_number == deployed_contract.w3.eth.get_block_number()

        # Calls should fail and caught with original web3 error
        try:
            _ = deployed_contract.functions.revertWithErrorOne().call()
            assert False, "Expected exception."
        except ContractCustomError as err:
            # Errors should still be caught with the original type.
            assert err.message == ErrorsContract.errors.One.selector

        # Test error handling with "transact"

        try:
            deployed_contract.functions.revertWithErrorOne().transact()
        except PypechainCallException as err:
            # Catch custom error
            assert isinstance(err.orig_exception, ContractCustomError)
            assert err.decoded_error == "One()"
            assert err.decoded_error_name == "One"
            assert err.decoded_error_args == ()
            assert err.contract_call_type == "transact"
            assert err.function_name == "revertWithErrorOne"
            assert err.fn_args == ()
            assert err.fn_kwargs == {}
            # We default the block number to be the "pending" block.
            # Since this transaction failed, and the chain is on "mine to transact"
            # mode, we expect the block number to be the next block.
            assert err.block_number == deployed_contract.w3.eth.get_block_number() + 1

        try:
            deployed_contract.functions.revertWithErrorOne().transact()
        except ContractCustomError as err:
            # Errors should still be caught with the original type.
            assert err.message == ErrorsContract.errors.One.selector

    def test_error_two(self, w3):
        """Test that we can decode strings."""
        deployed_contract = ErrorsContract.deploy(w3=w3, account=w3.eth.accounts[0])

        # Test error handling with "call"

        # Calls should fail and caught with custom error
        try:
            _ = deployed_contract.functions.revertWithErrorTwo().call()
            assert False, "Expected exception."
        except PypechainCallException as err:
            # Catch custom error
            assert isinstance(err.orig_exception, ContractCustomError)
            assert err.decoded_error == (
                "Two(I will not pledge allegiance to Bart., 0x0000000000000000000000000000000000000000, 255)"
            )
            assert err.decoded_error_name == "Two"
            # Args keep orig type
            assert err.decoded_error_args == (
                "I will not pledge allegiance to Bart.",
                "0x0000000000000000000000000000000000000000",
                255,
            )
            assert err.contract_call_type == "call"
            assert err.function_name == "revertWithErrorTwo"
            assert err.fn_args == ()
            assert err.fn_kwargs == {}
            # Call functions don't write a block, so the block should be identical
            assert err.block_number == deployed_contract.w3.eth.get_block_number()

        # Calls should fail and caught with original web3 error
        try:
            _ = deployed_contract.functions.revertWithErrorTwo().call()
            assert False, "Expected exception."
        except ContractCustomError as err:
            # Errors should still be caught with the original type.
            assert err.message
            assert err.message[:10] == ErrorsContract.errors.Two.selector

        # Test error handling with "transact"

        # Calls should fail and caught with custom error
        try:
            _ = deployed_contract.functions.revertWithErrorTwo().transact()
            assert False, "Expected exception."
        except PypechainCallException as err:
            # Catch custom error
            assert isinstance(err.orig_exception, ContractCustomError)
            assert err.decoded_error == (
                "Two(I will not pledge allegiance to Bart., 0x0000000000000000000000000000000000000000, 255)"
            )
            assert err.decoded_error_name == "Two"
            # Args keep orig type
            assert err.decoded_error_args == (
                "I will not pledge allegiance to Bart.",
                "0x0000000000000000000000000000000000000000",
                255,
            )
            assert err.contract_call_type == "transact"
            assert err.function_name == "revertWithErrorTwo"
            assert err.fn_args == ()
            assert err.fn_kwargs == {}
            # We default the block number to be the "pending" block.
            # Since this transaction failed, and the chain is on "mine to transact"
            # mode, we expect the block number to be the next block.
            assert err.block_number == deployed_contract.w3.eth.get_block_number() + 1

        # Calls should fail and caught with original web3 error
        try:
            _ = deployed_contract.functions.revertWithErrorTwo().transact()
            assert False, "Expected exception."
        except ContractCustomError as err:
            # Errors should still be caught with the original type.
            assert err.message
            assert err.message[:10] == ErrorsContract.errors.Two.selector

    def test_error_three(self, w3):
        """Test that we can decode structs and enums."""
        deployed_contract = ErrorsContract.deploy(w3=w3, account=w3.eth.accounts[0])

        in_struct = Ages(bart=1, lisa=2, homer=3, marge=4)

        # Test error handling with "call"

        # Calls should fail and caught with custom error
        try:
            _ = deployed_contract.functions.revertWithErrorThree(in_struct).call()
            assert False, "Expected exception."
        except PypechainCallException as err:
            # Catch custom error
            assert isinstance(err.orig_exception, ContractCustomError)
            assert err.decoded_error == ("Three(False, (1, 2, 3, 4), 0)")
            assert err.decoded_error_name == "Three"
            # Args keep orig type
            assert err.decoded_error_args == (
                False,
                (1, 2, 3, 4),  # TODO we may want to decode this into a struct
                0,
            )
            assert err.contract_call_type == "call"
            assert err.function_name == "revertWithErrorThree"
            assert err.fn_args == ((1, 2, 3, 4),)
            assert err.fn_kwargs == {}
            # Call functions don't write a block, so the block should be identical
            assert err.block_number == deployed_contract.w3.eth.get_block_number()

        # Calls should fail and caught with original web3 error
        try:
            _ = deployed_contract.functions.revertWithErrorThree(in_struct).call()
            assert False, "Expected exception."
        except ContractCustomError as err:
            # Errors should still be caught with the original type.
            assert err.message
            assert err.message[:10] == ErrorsContract.errors.Three.selector

        # Test error handling with "transact"

        # Calls should fail and caught with custom error
        try:
            _ = deployed_contract.functions.revertWithErrorThree(in_struct).transact()
            assert False, "Expected exception."
        except PypechainCallException as err:
            # Catch custom error
            assert isinstance(err.orig_exception, ContractCustomError)
            assert err.decoded_error == ("Three(False, (1, 2, 3, 4), 0)")
            assert err.decoded_error_name == "Three"
            # Args keep orig type
            assert err.decoded_error_args == (
                False,
                (1, 2, 3, 4),  # TODO we may want to decode this into a struct
                0,
            )
            assert err.contract_call_type == "transact"
            assert err.function_name == "revertWithErrorThree"
            assert err.fn_args == ((1, 2, 3, 4),)
            assert err.fn_kwargs == {}
            # We default the block number to be the "pending" block.
            # Since this transaction failed, and the chain is on "mine to transact"
            # mode, we expect the block number to be the next block.
            assert err.block_number == deployed_contract.w3.eth.get_block_number() + 1

        # Calls should fail and caught with original web3 error
        try:
            _ = deployed_contract.functions.revertWithErrorThree(in_struct).transact()
            assert False, "Expected exception."
        except ContractCustomError as err:
            # Errors should still be caught with the original type.
            assert err.message
            assert err.message[:10] == ErrorsContract.errors.Three.selector



"""A typed version of combomethod from eth_utils."""

import functools
from typing import Any, Callable, Concatenate, Generic, Optional, ParamSpec, Type, TypeVar

# TODO remove this file once https://github.com/ethereum/eth-utils/pull/264 gets merged

# We use generics to define the structure of the wrapped method
# Here, `T` is the `self` or `cls` object, `P` is the parameter of the wrapped function, and
# `R` is the return type

T = TypeVar("T")
P = ParamSpec("P")
R = TypeVar("R")


# We define the generic that attaches to the function we're decorating
# so here, P and R are the types of the parameters and return of the decorated function
class combomethod_typed(Generic[P, R]):  # pylint: disable=invalid-name
    """A typed version of combomethod from eth_utils."""

    # The callable `Any` takes place of the obj or class
    method: Callable[Concatenate[Any, P], R]

    # The method passed in has a spot for obj or class in `Any`
    def __init__(self, method: Callable[Concatenate[Any, P], R]) -> None:
        self.method = method

    # The getter allows for logic to call either cls or obj method
    # with the original type decorators in the output wrapper function
    def __get__(self, obj: Optional[T] = None, objtype: Optional[Type[T]] = None) -> Callable[P, R]:
        # The _wrapper function is unchanged from eth-utils
        @functools.wraps(self.method)
        def _wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            if obj is not None:
                return self.method(obj, *args, **kwargs)
            return self.method(objtype, *args, **kwargs)

        return _wrapper



"""Core pypechain functions used by generated files"""

from .base_event import BaseEvent, BaseEventArgs
from .combomethod_typed import combomethod_typed
from .contract_call_exception import (
    FailedTransaction,
    PypechainCallException,
    check_txn_receipt,
    handle_contract_logic_error,
)
from .contract_function import PypechainContractFunction
from .error import ErrorInfo, ErrorParams, PypechainBaseContractErrors, PypechainBaseError
from .utilities import (
    dataclass_to_tuple,
    expand_struct_type_str,
    get_abi_input_types,
    get_arg_type_names,
    rename_returned_types,
    tuple_to_dataclass,
)



"""Tests for overloading methods."""

from __future__ import annotations

import os
from pathlib import Path

import pytest
from web3 import Web3

from .types.ContractA import StructsA
from .types.ContractB import InnerStruct, StructsB
from .types.ContractC import ContractCContract

# using pytest fixtures necessitates this.
# pylint: disable=redefined-outer-name

current_path = os.path.abspath(os.path.dirname(__file__))
project_root = os.path.dirname(os.path.dirname(current_path))


@pytest.fixture
def deployed_contracts(w3: Web3) -> ContractCContract:
    """Deploys a StructsA contract and StructsB."""
    contract = ContractCContract.deploy(w3=w3, account=w3.eth.accounts[0])
    return contract


class TestStructs:
    """Tests pipeline from bots making trades to viewing the trades in the db"""

    def test_correct_files_exit(self):
        """StructsA and StructsB both inherit from IStructs.  StructsA also declares it's own
        struct.  Therefore we should expect to see types files for IStructs and StructsA, but not
        for StructsB since it doesn't declare its own.  IStructs is an interface file though, there
        is no ABI for it, so we shouldn't create a contract file."""

        expected_files = [
            "types/ContractA/__init__.py",
            "types/ContractA/ContractATypes.py",
            "types/ContractB/__init__.py",
            "types/ContractB/ContractBTypes.py",
            "types/ContractC/__init__.py",
            "types/ContractC/ContractCContract.py",
        ]

        results: list[tuple[bool, Path]] = []
        for file_path in expected_files:
            full_path = Path(current_path) / file_path
            results.append((full_path.exists(), full_path))

        failed_tests = [full_path for exists, full_path in results if not exists]

        assert len(failed_tests) == 0, f"Several files don't exist: {failed_tests=}"

    def test_build_struct(self, deployed_contracts: ContractCContract):
        """Test calling a method that returns the nested struct."""
        # Get the nested struct
        out_struct = deployed_contracts.functions.buildStruct().call()
        assert isinstance(out_struct, StructsB)
        assert isinstance(out_struct.structA, StructsA)
        assert isinstance(out_struct.structB, InnerStruct)
        assert out_struct.structA.intVal == 1
        assert out_struct.structA.strVal == "a"
        assert out_struct.structB.intVal == 2
        assert out_struct.structB.strVal == "b"



"""Tests for overloading methods."""

from __future__ import annotations

import os

from web3 import Web3
from web3.logs import DISCARD

from .types.Events import EventAEvent, EventBEvent, EventsContract

current_path = os.path.abspath(os.path.dirname(__file__))
project_root = os.path.dirname(os.path.dirname(current_path))


class TestEvents:
    """Tests events emitted from the contracts."""

    def test_process_receipt_typed(self, w3: Web3):
        """Test that we can use event filters."""
        deployed_contract = EventsContract.deploy(w3=w3, account=w3.eth.accounts[0])
        hash_a = deployed_contract.functions.emitOneEvent(0, "0x0000000000000000000000000000000000000000").transact()
        receipt_a = w3.eth.get_transaction_receipt(hash_a)
        hash_b = deployed_contract.functions.emitTwoEvents(1, "0x0000000000000000000000000000000000000001").transact()
        receipt_b = w3.eth.get_transaction_receipt(hash_b)

        # Web3 requires instantiated events (i.e., `EventA()`) for process_receipts
        # NOTE uses of `process_receipts` with an actual event will throw warnings
        # if the receipt has multiple events. This is expected, as web3 under the hood
        # loops through all events, regardless of the event attached. Hence, we discard
        # any events, as we handle looping over known events here.
        event_aa = list(deployed_contract.events.EventA().process_receipt_typed(receipt_a, errors=DISCARD))
        event_ab = list(deployed_contract.events.EventA().process_receipt_typed(receipt_b, errors=DISCARD))
        event_ba = list(deployed_contract.events.EventB().process_receipt_typed(receipt_a, errors=DISCARD))
        event_bb = list(deployed_contract.events.EventB().process_receipt_typed(receipt_b, errors=DISCARD))

        assert len(event_aa) == 1
        assert isinstance(event_aa[0], EventAEvent)
        assert event_aa[0].args.who == "0x0000000000000000000000000000000000000000"
        assert event_aa[0].args.value == 0

        assert len(event_ab) == 1
        assert isinstance(event_ab[0], EventAEvent)
        assert event_ab[0].args.who == "0x0000000000000000000000000000000000000001"
        assert event_ab[0].args.value == 1

        assert len(event_ba) == 0

        assert len(event_bb) == 1
        assert isinstance(event_bb[0], EventBEvent)

    def test_get_logs_typed(self, w3):
        """Test that we can get logs and the return is the type we expect."""
        deployed_contract = EventsContract.deploy(w3=w3, account=w3.eth.accounts[0])
        deployed_contract.functions.emitOneEvent(0, "0x0000000000000000000000000000000000000000").transact()
        deployed_contract.functions.emitTwoEvents(1, "0x0000000000000000000000000000000000000000").transact()

        event_a_logs = list(
            deployed_contract.events.EventA.get_logs_typed(
                argument_filters={"who": "0x0000000000000000000000000000000000000000"},
                from_block=0,
            )
        )
        event_b_logs = list(
            deployed_contract.events.EventB.get_logs_typed(
                from_block=0,
            )
        )

        assert len(event_a_logs) == 2
        assert isinstance(event_a_logs[0], EventAEvent)
        assert event_a_logs[0].args.who == "0x0000000000000000000000000000000000000000"
        assert event_a_logs[0].args.value == 0
        assert isinstance(event_a_logs[1], EventAEvent)
        assert event_a_logs[1].args.who == "0x0000000000000000000000000000000000000000"
        assert event_a_logs[1].args.value == 1

        assert len(list(event_b_logs)) == 1
        assert isinstance(event_b_logs[0], EventBEvent)

    def test_create_filter(self, w3):
        """Test that we can use event filters."""
        deployed_contract = EventsContract.deploy(w3=w3, account=w3.eth.accounts[0])
        tx_hash = deployed_contract.functions.emitOneEvent(0, "0x0000000000000000000000000000000000000000").transact()
        w3.eth.wait_for_transaction_receipt(tx_hash)

        event_a_filter = deployed_contract.events.EventA.create_filter(
            argument_filters={"who": "0x0000000000000000000000000000000000000000"},
            from_block=0,
        )
        event_b_filter = deployed_contract.events.EventB.create_filter(
            from_block=0,
        )

        a_events = event_a_filter.get_new_entries()
        b_events = event_b_filter.get_new_entries()

        assert len(a_events) == 1
        assert len(b_events) == 0

        tx_hash = deployed_contract.functions.emitTwoEvents(0, "0x0000000000000000000000000000000000000000").transact()
        w3.eth.wait_for_transaction_receipt(tx_hash)

        a_events = event_a_filter.get_new_entries()
        b_events = event_b_filter.get_new_entries()

        assert len(a_events) == 1
        assert len(b_events) == 1



"""Functions to render Python files from an abi usng a jinja2 template."""

from __future__ import annotations

import os
from functools import partial
from multiprocessing import Pool
from pathlib import Path

from pypechain.render.contract import ContractInfo, get_contract_infos, render_contract_file
from pypechain.render.init import render_init_file
from pypechain.render.types import render_types_file
from pypechain.utilities.abi import AbiInfo
from pypechain.utilities.file import write_string_to_file
from pypechain.utilities.format import format_file
from pypechain.utilities.types import RenderOutput


def render_contracts(
    abi_infos: list[AbiInfo],
    output_dir: str,
    line_length: int = 120,
    apply_formatting: bool = True,
    parallel: bool = False,
    chunksize: int | None = None,
) -> list[RenderOutput]:
    """Processes a single JSON file to generate class and types files.


    Parameters
    ----------
    abi_infos : list[AbiInfo]
        A list of objects containing each ABI and metadata.
    output_dir : str
        The directory where files will be written to.
    line_length : int, optional
        Black's line-length config option.
    apply_formatting : bool, optional
        If True, autoflake, isort and black will be applied to the file in that order, by default True.
    parallel : bool, optional
        If True, multiple processes will be used to process the ABIs in parallel.
    chunksize: int, optional
        If set, abis are processed in approximately groups of chunksize.
        Use 1 for one process per ABI.
        Defaults to the number of ABIs // 10.

    Returns
    -------
    list[RenderOutput]
        A list of filenames and definitions for the generated files.
    """
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments

    contract_infos = get_contract_infos(abi_infos)

    # this is what we are returning
    file_outputs: list[RenderOutput] = []

    # for every [ContractName] generate a:
    #    1. [ContractName]Contract.py
    #    2. [ContractName]Types.py
    #  The Contract file defines the Contract class and functions.
    #  The Types file has the structs and events defined in the solidity contract.

    if parallel:
        # make a parallel pool; defaults to the number of available CPUs
        with Pool() as pool:
            # chunksize should be increased with the number of iterables
            chunksize = len(contract_infos) // 10
            for file_output_sublist in pool.imap(
                partial(
                    render_contract, output_dir=output_dir, line_length=line_length, apply_formatting=apply_formatting
                ),
                list(contract_infos.values()),
                chunksize=chunksize,
            ):
                file_outputs.extend(file_output_sublist)
    # Sequential mode
    else:
        for contract_info in contract_infos.values():
            file_outputs.extend(
                render_contract(
                    contract_info=contract_info,
                    output_dir=output_dir,
                    line_length=line_length,
                    apply_formatting=apply_formatting,
                )
            )

    return file_outputs


def render_contract(
    contract_info: ContractInfo, output_dir: str, line_length: int, apply_formatting: bool
) -> list[RenderOutput]:
    """Get the file name for a single ContractInfo object


    Parameters
    ----------
    contract_info : ContractInfo
        A ContractInfo object created from an ABI.
    output_dir : str
        The directory where files will be written to.
    line_length : int, optional
        Black's line-length config option.
    apply_formatting : bool, optional
        If True, autoflake, isort and black will be applied to the file in that order, by default True.

    Returns
    -------
    list[RenderOutput]
        A list of filenames and definitions for the generated Contract and Types files.
    """
    file_outputs: list[RenderOutput] = []
    top_level_file_outputs: list[RenderOutput] = []
    contract_dir = Path(output_dir) / contract_info.contract_name
    os.makedirs(contract_dir)
    file_path = Path(contract_dir)

    rendered_contract_code = render_contract_file(contract_info)
    if rendered_contract_code:
        contract_file_name = contract_info.contract_name + "Contract.py"
        contract_file_path = Path(os.path.join(file_path, contract_file_name))
        write_string_to_file(contract_file_path, rendered_contract_code)
        if apply_formatting is True:
            format_file(contract_file_path, line_length)
        # We expose only the contract function from the contract file
        file_outputs.append(
            RenderOutput(
                filename=f"{contract_info.contract_name}Contract",
                definitions=[f"{contract_info.contract_name}Contract"],
            )
        )
        # We also build the top level __init__.py file
        top_level_file_outputs.append(
            RenderOutput(
                filename=f"{contract_info.contract_name}",
                definitions=[f"{contract_info.contract_name}Contract"],
            )
        )

    rendered_types_code = render_types_file(contract_info)
    if rendered_types_code:
        types_file_name = contract_info.contract_name + "Types.py"
        types_file_path = Path(os.path.join(file_path, types_file_name))
        write_string_to_file(types_file_path, rendered_types_code)
        if apply_formatting is True:
            format_file(types_file_path, line_length)
        # Expose events and structs in init.
        # TODO errors also exists in the types file, but isn't used.
        # We avoid exposing to avoid confusion with error classes in
        # the contract file.
        file_outputs.append(
            RenderOutput(
                filename=f"{contract_info.contract_name}Types",
                definitions=(
                    [f"{struct.name}" for struct in contract_info.structs.values()]
                    + [f"{event.capitalized_name}Event" for event in contract_info.events.values()]
                ),
            )
        )

    render_init_file(contract_dir, file_outputs)
    if apply_formatting is True:
        format_file(contract_dir / "__init__.py", line_length, remove_unused_imports=False)

    return top_level_file_outputs



"""Tests for rendering the init file."""

import os

from pypechain.utilities.templates import get_jinja_env
from pypechain.utilities.types import RenderOutput

# using pytest fixtures necessitates this.
# pylint: disable=redefined-outer-name

current_path = os.path.abspath(os.path.dirname(__file__))
project_root = os.path.dirname(os.path.dirname(current_path))


class TestRenderInit:
    """Tests rendering for the init file using snapshots"""

    def test_render_init(self, snapshot):
        """
        Tests the code gen init file using snapshots
        """

        env = get_jinja_env()
        init_template = env.get_template("init.py.jinja2")

        # TODO: add return types to function calls

        file_outputs = [
            RenderOutput(filename="ContractFile1", definitions=["ContractFile1Contract"]),
            RenderOutput(filename="ContractFile2", definitions=["ContractFile2Contract"]),
            RenderOutput(filename="TypesFile", definitions=["Struct1"]),
            RenderOutput(filename="TypesFile", definitions=["Struct2"]),
        ]

        # To avoid having to regenerate snapshots per version bump, we hard code the test version to 0.0.0 here
        init_code = init_template.render(
            pypechain_version="_version_ignored",
            file_outputs=file_outputs,
        )

        snapshot.snapshot_dir = "snapshots"  # This line is optional.
        snapshot.assert_match(init_code, "expected_init.py")



"""Functions to render __init__.py from a list of filenames usng a jinja2 template."""

import importlib.metadata
from pathlib import Path

from pypechain.utilities.file import write_string_to_file
from pypechain.utilities.templates import get_jinja_env
from pypechain.utilities.types import RenderOutput


def render_init_file(output_dir: Path, file_outputs: list[RenderOutput]):
    """Creates an __init__.py file that imports all other files."""
    env = get_jinja_env()
    init_template = env.get_template("init.py.jinja2")
    init_code = init_template.render(
        pypechain_version=importlib.metadata.version("pypechain"),
        file_outputs=file_outputs,
    )
    init_file_path = Path(f"{output_dir}/__init__.py")
    write_string_to_file(init_file_path, init_code)



"""Script to generate typed web3.py classes for Solidity contracts."""

from __future__ import annotations

import argparse
import importlib.metadata
import os
import shutil
import sys
from pathlib import Path
from typing import NamedTuple, Sequence

from pypechain.render import render_contracts
from pypechain.render.init import render_init_file
from pypechain.utilities.abi import AbiInfo, load_abi_infos_from_file
from pypechain.utilities.format import format_file


def main(argv: Sequence[str] | None = None) -> None:
    """Main entrypoint for pypechain cli.

    Parameters
    ----------
    argv : Sequence[str] | None, optional
        Command line arguments
    """
    abi_file_path, output_dir, line_length, apply_formatting, parallel = parse_arguments(argv)
    pypechain(abi_file_path, output_dir, line_length, apply_formatting, parallel)


def pypechain(
    abi_file_path: str,
    output_dir: str = "pypechain_types",
    line_length: int = 120,
    apply_formatting: bool = True,
    parallel: bool = False,
):
    """Generate class files for a given abi.

    Parameters
    ----------
    abi_file_path : str
        Path to the abi JSON file.
    output_dir: str, optional
        Path to the directory where files will be generated. Defaults to 'pypechain_types'.
    line_length : int, optional
        Optional argument for the output file's maximum line length. Defaults to 120.
    apply_formatting : bool, optional
        If True, autoflake, isort and black will be applied to the file in that order, by default True.
    """
    # List to store all JSON ABI files to be processed
    json_files_to_process: list[Path] = []

    # Check if provided path is a directory or file
    if os.path.isdir(abi_file_path):
        # If directory, gather all JSON files recursively in the directory
        json_files_to_process.extend(gather_json_files(abi_file_path))
    else:
        # Otherwise, add the single file to the list
        json_files_to_process.append(Path(abi_file_path))

    # Parse the files and gather AbiInfos
    abi_infos: list[AbiInfo] = []
    for json_file in json_files_to_process:
        # Skip failing files
        try:
            infos = load_abi_infos_from_file(json_file)
            abi_infos.extend(infos)
        except Exception as e:  # pylint: disable=broad-except
            print(f"Skipping {json_file} due to error {e}")

    # Create/clear the output directory
    setup_directory(output_dir)

    # Since setup directory looks for `pypechain.version`, we make this file first
    # Make a pypechain.version file
    with open(Path(output_dir) / "pypechain.version", "w", encoding="utf-8") as f:
        f.write(f"pypechain == {importlib.metadata.version('pypechain')}")

    # Now process all gathered files
    file_outputs = render_contracts(abi_infos, output_dir, line_length, apply_formatting, parallel=parallel)

    # Make the top level __init__.py file
    render_init_file(Path(output_dir), file_outputs)
    if apply_formatting is True:
        format_file(Path(output_dir) / "__init__.py", line_length, remove_unused_imports=False)


def gather_json_files(directory: str) -> list:
    """Gather all JSON files in the specified directory and its subdirectories."""
    return list(Path(directory).rglob("*.json"))


def setup_directory(directory: str) -> None:
    """Set up the output directory. If it exists, clear it. Otherwise, create it."""

    print(f"Setting up output directory {directory}.")

    # If the directory exists, make sure it's a pypechain output directory
    if os.path.exists(directory):
        if os.path.isfile(f"{directory}/pypechain.version"):
            # Clear the directory if it's a pypechain output directory
            shutil.rmtree(directory)
        else:
            # Print error if directory isn't empty
            if not len(os.listdir(directory)) == 0:
                raise ValueError(f"Directory {directory} isn't empty, and is not a pypechain output directory.")

    # Create the directory
    os.makedirs(directory, exist_ok=True)


class Args(NamedTuple):
    """Command line arguments for pypechain."""

    abi_file_path: str
    output_dir: str
    line_length: int
    apply_formatting: bool
    parallel: bool = False


def namespace_to_args(namespace: argparse.Namespace) -> Args:
    """Convert argprase.Namespace to Args."""
    return Args(
        abi_file_path=namespace.abi_file_path,
        output_dir=namespace.output_dir,
        line_length=namespace.line_length,
        apply_formatting=namespace.apply_formatting,
        parallel=namespace.parallel,
    )


def parse_arguments(argv: Sequence[str] | None = None) -> Args:
    """Parse input arguments."""
    parser = argparse.ArgumentParser(description="Generates class files for a given abi.")

    parser.add_argument(
        "abi_file_path",
        help="Path to the abi JSON file or directory containing multiple JSON files.",
    )

    parser.add_argument(
        "--output-dir",
        default="pypechain_types",
        help="Path to the directory where files will be generated. Defaults to pypechain_types.",
    )
    parser.add_argument(
        "--line-length",
        type=int,
        default=120,
        help="Optional argument for the output file's maximum line length. Defaults to 120.",
    )
    parser.add_argument(
        "--apply-formatting",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="Optional argument to apply formatting to each file. Defaults to True.",
    )
    parser.add_argument(
        "--parallel",
        default=False,
        action="store_true",
        help="Flag controlling building in parallel. Defaults to False",
    )

    # Use system arguments if none were passed
    if argv is None:
        argv = sys.argv

    # If no arguments were passed, display the help message and exit
    if len(argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    return namespace_to_args(parser.parse_args())


if __name__ == "__main__":
    main()



"""Tests for deploying contracts with linked libraries methods."""

from __future__ import annotations

import os

import pytest
from eth_account.account import Account
from eth_account.signers.local import LocalAccount
from web3 import Web3
from web3.types import RPCEndpoint

from .types.Contract import ContractContract
from .types.MyLibrary import MyLibraryContract

# using pytest fixtures necessitates this.
# pylint: disable=redefined-outer-name
current_path = os.path.abspath(os.path.dirname(__file__))
project_root = os.path.dirname(os.path.dirname(current_path))


@pytest.fixture
def local_account(w3) -> LocalAccount:
    """Generates a random local account.

    Returns
    -------
    LocalAccount
        An account with a random private key and therefore public key.
    """
    local_account = Account.create("")

    amount_wei = int(100 * 1e18)
    params = [local_account.address, hex(amount_wei)]  # account, amount
    w3.provider.make_request(method=RPCEndpoint("anvil_setBalance"), params=params)

    return local_account


@pytest.mark.usefixtures("process_contracts")
class TestDeployLinking:
    """Tests deployment methods for contracts with no constructor"""

    def test_deployment_requires_contract_linking(self, w3: Web3, local_account: LocalAccount):
        """Tests contract deployment fails if contract linking is not provided."""
        # This should fail: Contract requires linked references
        with pytest.raises(TypeError):
            _ = ContractContract.deploy(w3, local_account)  # type: ignore # pylint: disable=no-value-for-parameter

    def test_deployment_with_contract_linking(self, w3: Web3, local_account: LocalAccount):
        """Tests deployment using contract linking in deploy."""

        # Deploy the library first
        library_contract = MyLibraryContract.deploy(w3, local_account)
        contract = ContractContract.deploy(
            w3,
            local_account,
            link_references=ContractContract.LinkReferences(
                MyLibrary=library_contract,
            ),
        )

        assert contract.functions.add(1, 2).call() == 3



"""Dataclasses for all structs in the Errors contract.

DO NOT EDIT.  This file was generated by pypechain v0.0.46.
See documentation at https://github.com/delvtech/pypechain """

# super() call methods are generic, while our version adds values & types
# pylint: disable=arguments-differ

# contracts have PascalCase names
# pylint: disable=invalid-name

# contracts control how many attributes and arguments we have in generated code
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments

# unable to determine which imports will be used in the generated code
# pylint: disable=unused-import

# we don't need else statement if the other conditionals all have return,
# but it's easier to generate
# pylint: disable=no-else-return

# We import this contract itself to ensure all nested structs have a fully qualified name.
# We use this to avoid namespace collisions, as well as having a uniform
# type structure to do lookups when functions return these structs.
# pylint: disable=import-self


from __future__ import annotations

from dataclasses import dataclass

from pypechain.core import ErrorInfo, ErrorParams


@dataclass
class Ages:
    """Ages struct."""

    bart: int
    lisa: int
    homer: int
    marge: int


OneError = ErrorInfo(
    inputs=[],
    name="One",
    selector="0xbe0c2110",
    signature="One()",
)

ThreeError = ErrorInfo(
    inputs=[
        ErrorParams(name="trueOrFalse", python_type="bool", solidity_type="bool"),
        ErrorParams(name="theSimpsons", python_type="tuple", solidity_type="tuple"),
        ErrorParams(name="who", python_type="int", solidity_type="uint8"),
    ],
    name="Three",
    selector="0x09b8b989",
    signature="Three(bool,(uint256,uint256,uint256,uint256),uint8)",
)

TwoError = ErrorInfo(
    inputs=[
        ErrorParams(name="message", python_type="str", solidity_type="string"),
        ErrorParams(name="who", python_type="str", solidity_type="address"),
        ErrorParams(name="value", python_type="int", solidity_type="uint8"),
    ],
    name="Two",
    selector="0x01e3e2f6",
    signature="Two(string,address,uint8)",
)



"""Tests for overloading methods."""

from __future__ import annotations

import os

import pytest
from web3 import Web3

from .types.ReturnTypes import InnerStruct, NestedStruct, ReturnTypesContract, SimpleStruct

# using pytest fixtures necessitates this.
# pylint: disable=redefined-outer-name

current_path = os.path.abspath(os.path.dirname(__file__))
project_root = os.path.dirname(os.path.dirname(current_path))


@pytest.fixture
def deployed_contract(w3: Web3):
    """Deploys a ReturnTypes contract."""
    return ReturnTypesContract.deploy(w3=w3, account=w3.eth.accounts[0])


@pytest.mark.usefixtures("process_contracts")
class TestReturnTypes:
    """Tests pipeline from bots making trades to viewing the trades in the db"""

    def test_no_name_single_value(self, deployed_contract: ReturnTypesContract):
        """Tests single value"""
        result: int = deployed_contract.functions.noNameSingleValue(1).call()
        assert result == 1

    def test_no_name_two_values(self, deployed_contract: ReturnTypesContract):
        """Tests two values"""
        func = deployed_contract.functions.noNameTwoValues("a string")
        result = func.call()
        assert isinstance(result, func.ReturnValues)
        assert result == func.ReturnValues("a string", 2)

    def test_named_single_value(self, deployed_contract: ReturnTypesContract):
        """Tests named single value"""
        result: int = deployed_contract.functions.namedSingleValue(1, 2).call()
        assert result == 3

    def test_named_two_values(self, deployed_contract: ReturnTypesContract):
        """Tests two named values"""
        func = deployed_contract.functions.namedTwoValues(1, 2)
        result = func.call()
        assert isinstance(result, func.ReturnValues)
        assert result == func.ReturnValues(2, 1)

    def test_single_simple_struct(self, deployed_contract: ReturnTypesContract):
        """Tests a struct"""
        result = deployed_contract.functions.singleSimpleStruct().call()
        assert result == SimpleStruct(1, "You are number 1")

    def test_single_nested_struct(self, deployed_contract: ReturnTypesContract):
        """Tests a nested struct"""
        result = deployed_contract.functions.singleNestedStruct().call()
        assert result == NestedStruct(1, "You are number 1", InnerStruct(True))

    def test_two_simple_structs(self, deployed_contract: ReturnTypesContract):
        """Tests two structs"""
        func = deployed_contract.functions.twoSimpleStructs()
        result = func.call()
        assert isinstance(result, func.ReturnValues)
        assert result == func.ReturnValues(SimpleStruct(1, "You are number 1"), SimpleStruct(2, "You are number 2"))

    def test_two_mixed_structs(self, deployed_contract: ReturnTypesContract):
        """Tests two structs, one nested"""
        func = deployed_contract.functions.twoMixedStructs()
        result = func.call()
        assert isinstance(result, func.ReturnValues)
        assert result == func.ReturnValues(
            SimpleStruct(1, "You are number 1"), NestedStruct(2, "You are number 2", InnerStruct(True))
        )

    def test_named_single_struct(self, deployed_contract: ReturnTypesContract):
        """Tests a named struct"""
        result = deployed_contract.functions.namedSingleStruct().call()
        assert result == SimpleStruct(1, "You are number 1")

    def test_named_two_mixed_structs(self, deployed_contract: ReturnTypesContract):
        """Tests two named structs, one nested"""
        func = deployed_contract.functions.namedTwoMixedStructs()
        result = func.call()
        assert isinstance(result, func.ReturnValues)
        assert result == func.ReturnValues(
            SimpleStruct(1, "You are number 1"), NestedStruct(2, "You are number 2", InnerStruct(True))
        )

    def test_mix_structs_and_primitives(self, deployed_contract: ReturnTypesContract):
        """Tests two structs, one nested, and other values returned"""
        func = deployed_contract.functions.mixStructsAndPrimitives()

        result = func.call()

        assert isinstance(result, func.ReturnValues)
        assert result == (
            SimpleStruct(1, "You are number 1"),
            NestedStruct(2, "You are number 2", InnerStruct(True)),
            1,
            "ReturnTypesContract",
            False,
        )

    def test_list_of_dataclass(self, deployed_contract: ReturnTypesContract):
        """Tests a function that returns list[Struct]."""
        func = deployed_contract.functions.singleNestedStructArray()

        result = func.call()

        assert len(result) == 2
        assert isinstance(result[0], NestedStruct)
        assert isinstance(result[1], NestedStruct)
        assert result[0] == NestedStruct(1, "You are number 1", InnerStruct(True))
        assert result[1] == NestedStruct(2, "You are number 2", InnerStruct(True))



"""Tests for sign and transact."""

from __future__ import annotations

import os

import pytest
from eth_account.account import Account
from eth_account.signers.local import LocalAccount
from eth_utils.conversions import to_bytes
from eth_utils.crypto import keccak
from eth_utils.curried import text_if_str
from hexbytes import HexBytes
from web3 import Web3
from web3.types import RPCEndpoint

from pypechain.core import PypechainCallException

from .types.Transact import TransactContract

# using pytest fixtures necessitates this.
# pylint: disable=redefined-outer-name


def make_private_key(extra_entropy: str = "SOME STRING") -> str:
    """Make a private key.

    Arguments
    ---------
    extra_entropy: str, optional
        Any string used to add entropy to the keccak hash.
        Defaults to "SOME STRING".

    Returns
    -------
    str
        The private key.
    """
    extra_key_bytes = text_if_str(to_bytes, extra_entropy)
    key_bytes = keccak(os.urandom(32) + extra_key_bytes)
    return HexBytes(key_bytes).hex()


@pytest.mark.usefixtures("process_contracts")
class TestStructs:
    """Tests pipeline from bots making trades to viewing the trades in the db"""

    def test_sign_and_transact(self, w3: Web3):
        """Test the sign and transact function"""
        deployed_contract = TransactContract.deploy(w3=w3, account=w3.eth.accounts[0])

        # Make a new account that's not a default anvil account
        account: LocalAccount = Account().from_key(make_private_key())

        # Fund the account with eth
        w3.provider.make_request(method=RPCEndpoint("anvil_setBalance"), params=[account.address, hex(int(1e18))])

        # Expect failure with transaction without signing
        with pytest.raises(PypechainCallException):
            _ = deployed_contract.functions.TransactFunction().transact({"from": account.address})

        # Expect success with sign and transact
        tx_hash = deployed_contract.functions.TransactFunction().sign_and_transact(account=account)
        receipt = w3.eth.get_transaction_receipt(tx_hash)
        out_events = deployed_contract.events.Success().process_receipt_typed(receipt)
        assert len(list(out_events)) == 1

    def test_sign_transact_and_wait(self, w3: Web3):
        """Test the sign and transact function"""
        deployed_contract = TransactContract.deploy(w3=w3, account=w3.eth.accounts[0])

        # Make a new account that's not a default anvil account
        account: LocalAccount = Account().from_key(make_private_key())

        # Fund the account with eth
        w3.provider.make_request(method=RPCEndpoint("anvil_setBalance"), params=[account.address, hex(int(1e18))])

        # Expect failure with transaction without signing
        with pytest.raises(PypechainCallException):
            _ = deployed_contract.functions.TransactFunction().transact({"from": account.address})

        # Expect success with sign and transact
        receipt = deployed_contract.functions.TransactFunction().sign_transact_and_wait(
            account=account, validate_transaction=True
        )
        out_events = deployed_contract.events.Success().process_receipt_typed(receipt)
        assert len(list(out_events)) == 1



"""Utility functions for the pypechain generated files."""

from __future__ import annotations

from dataclasses import fields, is_dataclass
from typing import Any, Iterable, Tuple, TypeVar, cast, get_args

from eth_typing import ABIFunction
from eth_utils.abi import collapse_if_tuple

T = TypeVar("T")


def tuple_to_dataclass(cls: type[T], structs: dict[str, Any], tuple_data: Any | Tuple[Any, ...]) -> T:
    """
    Converts a tuple (including nested tuples) to a dataclass instance.  If cls is not a dataclass,
    then the data will just be passed through this function.

    Parameters
    ----------
    cls: type[T]
        The dataclass type to which the tuple data is to be converted.
    structs: dict[str, Any]
        Mapping from a struct name (key) to the run-time type definition (value).
    tuple_data: Any | Tuple[Any, ...]
        A tuple (or nested tuple) of values to convert into a dataclass instance.

    Returns
    -------
    T
        Either an instance of cls populated with data from tuple_data or tuple_data itself.
    """
    if not is_dataclass(cls):
        return cast(T, tuple_data)

    field_types = {field.name: field.type for field in fields(cls)}
    field_values = {}

    for (field_name, field_type), value in zip(field_types.items(), tuple_data):
        if field_type in structs:
            field_type = structs[field_type]
        if is_dataclass(field_type):
            # Type narrowing
            assert isinstance(field_type, type)
            # Recursively convert nested tuples to nested dataclasses
            field_values[field_name] = tuple_to_dataclass(field_type, structs, value)
        elif isinstance(value, tuple) and not getattr(field_type, "_name", None) == "Tuple":
            # If it's a tuple and the field is not intended to be a tuple, assume it's a nested dataclass
            # Type narrowing
            assert isinstance(field_type, type)
            field_values[field_name] = tuple_to_dataclass(field_type, structs, value)
        else:
            # Otherwise, set the primitive value directly
            field_values[field_name] = value

    return cls(**field_values)


def dataclass_to_tuple(instance: Any) -> Any:
    """Convert a dataclass instance to a tuple, handling nested dataclasses.
    If the input is not a dataclass, return the original value.

    Parameters
    ----------
    instance : Any
        The dataclass instance to convert to a tuple.  If it is not it passes through.

    Returns
    -------
    Any
        either a tuple or the original value
    """
    # We expect vector inputs as lists, as set by the type hint of
    # the function arguments.
    if isinstance(instance, list):
        return [dataclass_to_tuple(x) for x in instance]

    if not is_dataclass(instance):
        return instance

    def convert_value(value: Any) -> Any:
        """Convert nested dataclasses to tuples recursively, or return the original value."""
        if is_dataclass(value):
            return dataclass_to_tuple(value)
        return value

    return tuple(convert_value(getattr(instance, field.name)) for field in fields(instance))


def get_arg_type_names(in_args: tuple[Any, ...] | list[Any] | Any) -> str:
    """Returns the types of an input sequence of arguments recursively.

    Parameters
    ----------
    in_args : tuple[Any, ...] | list[Any] | Any
        The input sequence of arguments.

    Returns
    -------
    str
        The types of the input sequence of arguments.
    """
    if isinstance(in_args, tuple):
        out = f"({', '.join(get_arg_type_names(arg) for arg in in_args)})"
        return out
    if isinstance(in_args, list):
        # We only look at the first element of the list for types
        # if it's a list, as we expect a list input be mapped to
        # a vector input type in solidity, which must all be of
        # the same type. This is also ensured via type hinting in
        # the contract function.
        out = f"list[{get_arg_type_names(in_args[0])}]"
        return out
    return type(in_args).__name__


def expand_struct_type_str(in_type: tuple[str, ...] | str, structs: dict[str, Any]) -> str:
    """Given a type string, expand any structs to tuples."""
    if isinstance(in_type, str):
        # handle list[types]
        if in_type.startswith("list[") and in_type.endswith("]"):
            return f"list[{expand_struct_type_str(in_type[5:-1], structs)}]"

        if in_type in structs:
            # Expand struct types to tuples
            f = [expand_struct_type_str(str(field.type), structs) for field in fields(structs[in_type])]
            return f"({', '.join(f)})"
        return in_type

    if isinstance(in_type, tuple):
        return f"({', '.join(expand_struct_type_str(field, structs) for field in in_type)})"

    raise TypeError(f"Unsupported type: {in_type}")


def rename_returned_types(
    structs: dict[str, Any], return_types: list[Any] | Any, raw_values: list[str | int | tuple] | str | int | tuple
) -> tuple:
    """Convert structs in the return value to known dataclasses.

    Parameters
    ----------
    return_types : list[str] | str
        The type or list of types returned from a contract.
    raw_values : list[str  |  int | tuple] | str | int | tuple
        The actual returned values from the contract.

    Returns
    -------
    tuple
        The return types.
    """
    # cover case of multiple return values
    if isinstance(return_types, list):
        # Ensure raw_values is a tuple for consistency
        if not isinstance(raw_values, list):
            raw_values = (raw_values,)

        # Convert the tuple to the dataclass instance using the utility function
        converted_values = []
        for return_type, value in zip(return_types, raw_values):
            if type(return_type) == type(list[Any]):  # pylint: disable=unidiomatic-typecheck
                raise NotImplementedError("Multiple return values of type list[...] is not supported.")
            converted_values.append(tuple_to_dataclass(return_type, structs, value))
        converted_values = tuple(converted_values)

        return converted_values

    # cover case of single return type
    # single return type is a list of SomeType, aka `list[SomeType]`
    if type(return_types) == type(list[Any]):  # pylint: disable=unidiomatic-typecheck
        inner_types = get_args(return_types)
        # make sure there is only one inner type
        if len(inner_types) != 1:
            raise NotImplementedError("Only a single inner type in list[...] is supported")
        inner_type = inner_types[0]
        # make sure the type is not also a list
        if type(inner_type) == type(list[Any]):  # pylint: disable=unidiomatic-typecheck
            raise NotImplementedError("Type list[list[...]] is not supported.")
        # type narrowing
        assert isinstance(raw_values, Iterable)
        # loop over inner values & convert those to dataclasses
        converted_value = []
        for value in raw_values:
            converted_value.append(tuple_to_dataclass(inner_type, structs, value))
        return tuple(converted_value)

    # single return type is a standard type or dataclass
    converted_value = tuple_to_dataclass(return_types, structs, raw_values)
    return converted_value


def get_abi_input_types(abi: ABIFunction) -> list[str]:
    """Gets all the solidity input types for a function or error.

    Cribbed from web3._utils.abi.py file.

    Parameters
    ----------

    abi: ABIFunction
        The ABIFunction or ABIError that we want to get input types for.

    Returns
    -------
    list[str]
        A list of solidity input types.

    """

    if "inputs" not in abi and (abi.get("type") == "fallback" or abi.get("type") == "receive"):
        return []
    return [collapse_if_tuple(cast(dict[str, Any], arg)) for arg in abi.get("inputs", [])]
